# -*- coding: utf-8 -*-
"""
Created on Sat Aug  3 23:20:19 2019

@author: ydima
"""

import os


class BCPandasException(Exception):
    pass


class BCPandasValueError(BCPandasException):
    pass


# BCP terms
IN = "in"
OUT = "out"
QUERYOUT = "queryout"
TABLE = "table"
VIEW = "view"
QUERY = "query"

DIRECTIONS = (IN, OUT, QUERYOUT)
SQL_TYPES = (TABLE, VIEW, QUERY)
IF_EXISTS_OPTIONS = ("append", "replace", "fail")


# Text settings
_DELIMITER_OPTIONS = (",", "|", "\t")
QUOTECHAR = '"'
NEWLINE = os.linesep

# BCP Format File terms
SQLCHAR = "SQLCHAR"
sql_collation = "SQL_Latin1_General_CP1_CI_AS"


def get_delimiter(df):
    for delim in _DELIMITER_OPTIONS:
        if not df.applymap(lambda x: delim in x if isinstance(x, str) else False).any().any():
            return delim
    raise BCPandasValueError(
        f"Data contains all of the possible delimiter characters ({_DELIMITER_OPTIONS}), "
        "cannot use BCP to import it. Replace one of the possible delimiter characters in "
        "your data, or use another method besides bcpandas. \nFurther background: \n "
        "https://docs.microsoft.com/en-us/sql/relational-databases/import-export/specify-field-and-row-terminators-sql-server#characters-supported-as-terminators"
    )



from .main import to_sql, read_sql, SqlCreds
from .utils import bcp, sqlcmd

import subprocess
import warnings

name = "bcpandas"
__version__ = "0.1.0"


# BCP check
cmds = [["bcp", "-v"], ["sqlcmd", "-?"]]
for cmd in cmds:
    try:
        subprocess.run(cmd)
    except FileNotFoundError:
        warnings.warn(
            f"{cmd[0].upper()} utility not installed or not found in PATH, bcpandas will not work!"
        )

del subprocess, warnings, cmd



# -*- coding: utf-8 -*-
"""
Created on Sat Aug  3 23:07:15 2019

@author: ydima
"""

import csv
import logging
import os

import pandas as pd

from .constants import (
    BCPandasValueError,
    get_delimiter,
    _DELIMITER_OPTIONS,
    IF_EXISTS_OPTIONS,
    IN,
    NEWLINE,
    OUT,
    QUERYOUT,
    QUOTECHAR,
    SQL_TYPES,
    TABLE,
    VIEW,
    QUERY,
)
from .utils import _get_sql_create_statement, bcp, build_format_file, get_temp_file, sqlcmd

logger = logging.getLogger(__name__)


class SqlCreds:
    """
    Credential object for all SQL operations.

    If `username` and `password` are not provided, `with_krb_auth` will be `True`.

    Parameters
    ----------
    server : str
    database : str
    username : str, optional
    password : str, optional

    Returns
    -------
    `bcpandas.SqlCreds`
    """

    def __init__(self, server, database, username=None, password=None):
        if not server or not database:
            raise BCPandasValueError(
                f"Server and database can't be None, you passed {server}, {database}"
            )
        self.server = server
        self.database = database
        if username and password:
            self.username = username
            self.password = password
            self.with_krb_auth = False
        else:
            self.with_krb_auth = True
        logger.info(f"Created creds:\t{self}")

    @classmethod
    def from_engine(cls, engine):
        """
        Alternate constructor from a `sqlalchemy.engine.base.Engine` object.

        Alternate constructor, from a `sqlalchemy.engine.base.Engine` that uses `pyodbc` as the DBAPI 
        (which is the SQLAlchemy default for MS SQL) and using an exact PyODBC connection string (not DSN or hostname).
        See https://docs.sqlalchemy.org/en/13/dialects/mssql.html#connecting-to-pyodbc for more.
        
        Parameters
        ----------
        engine : `sqlalchemy.engine.base.Engine`
            The SQLAlchemy engine object, configured as described above

        Returns
        -------
        `bcpandas.SqlCreds`
        """
        try:
            # get the odbc url part from the engine, split by ';' delimiter
            conn_url = engine.url.query["odbc_connect"].split(";")
            # convert into dict
            conn_dict = {x.split("=")[0]: x.split("=")[1] for x in conn_url if "=" in x}

            return cls(
                server=conn_dict["Server"].replace("tcp:", "").replace(",1433", ""),
                database=conn_dict["Database"],
                username=conn_dict["UID"],
                password=conn_dict["PWD"],
            )
        except (KeyError, AttributeError):
            raise BCPandasValueError(
                "The supplied 'engine' object could not be parsed correctly, try creating a SqlCreds object manually."
            )

    def __repr__(self):
        # adopted from https://github.com/erdewit/ib_insync/blob/master/ib_insync/objects.py#L51
        clsName = self.__class__.__qualname__
        kwargs = ", ".join(f"{k}={v!r}" for k, v in self.__dict__.items() if k != "password")
        if hasattr(self, "password"):
            kwargs += ", password=[REDACTED]"
        return f"{clsName}({kwargs})"

    __str__ = __repr__


def to_sql(
    df,
    table_name,
    creds,
    sql_type="table",
    schema="dbo",
    index=True,
    if_exists="fail",
    batch_size=None,
    debug=False,
):
    """
    Writes the pandas DataFrame to a SQL table or view.

    Will write all columns to the table or view. 
    Assumes the SQL table or view has the same number, name, and type of columns.
    To only write parts of the DataFrame, filter it beforehand and pass that to this function.

    Parameters
    ----------
    df : pandas.DataFrame
    table_name : str
        Name of SQL table or view, without the schema
    creds : bcpandas.SqlCreds
        The credentials used in the SQL database.
    sql_type : {'table', 'view'}, default 'table'
        The type of SQL object of the destination.
    schema : str, default 'dbo'
        The SQL schema.
    index : bool, default True
        Write DataFrame index as a column. Uses the index name as the column
        name in the table.
    if_exists : {'fail', 'replace', 'append'}, default 'fail'
        How to behave if the table already exists.
        * fail: Raise a BCPandasValueError.
        * replace: Drop the table before inserting new values.
        * append: Insert new values to the existing table.
    batch_size : int, optional
        Rows will be written in batches of this size at a time. By default,
        all rows will be written at once.
    debug : bool, default False
        If True, will not delete the temporary CSV and format files, and will output their location.
    """
    # validation
    assert sql_type in SQL_TYPES
    assert if_exists in IF_EXISTS_OPTIONS

    delim = get_delimiter(df)

    # save to temp path
    csv_file_path = get_temp_file()
    df.to_csv(
        path_or_buf=csv_file_path,
        sep=delim,
        header=False,
        index=False,
        quoting=csv.QUOTE_MINIMAL,  # pandas default
        quotechar=QUOTECHAR,
        line_terminator=NEWLINE,
        doublequote=True,
        escapechar=None,  # not needed, as using doublequote
    )
    logger.debug(f"Saved dataframe to temp CSV file at {csv_file_path}")

    # build format file
    fmt_file_path = get_temp_file()
    fmt_file_txt = build_format_file(df=df, delimiter=delim)
    with open(fmt_file_path, "w") as ff:
        ff.write(fmt_file_txt)
    logger.debug(f"Created BCP format file at {fmt_file_path}")

    try:
        if if_exists == "fail":
            _qry = """SELECT * 
                 FROM INFORMATION_SCHEMA.{_typ}S 
                 WHERE TABLE_SCHEMA = '{_schema}' 
                 AND TABLE_NAME = '{_tbl}'"""
            res = sqlcmd(
                creds=creds,
                command=_qry.format(_typ=sql_type.upper(), _schema=schema, _tbl=table_name),
            )
            if res is not None:
                raise BCPandasValueError(
                    f"The {sql_type} called {schema}.{table_name} already exists, "
                    f"`if_exists` param was set to `fail`."
                )
        elif if_exists == "replace":
            sqlcmd(
                creds=creds,
                command=_get_sql_create_statement(df=df, table_name=table_name, schema=schema),
            )
        elif if_exists == "append":
            pass  # don't need to do anything

        # either way, BCP data in
        bcp(
            sql_item=table_name,
            direction=IN,
            flat_file=csv_file_path,
            format_file_path=fmt_file_path,
            creds=creds,
            sql_type=sql_type,
            schema=schema,
            batch_size=batch_size,
        )
    finally:
        if not debug:
            logger.debug(f"Deleting temp CSV and format files")
            os.remove(csv_file_path)
            os.remove(fmt_file_path)
        else:
            logger.debug(
                f"`to_sql` DEBUG mode, not deleting the files. CSV file is at "
                f"{csv_file_path}, format file is at {fmt_file_path}"
            )


def read_sql(table_name, creds, sql_type="table", schema="dbo", batch_size=None, debug=False):
    """
    Reads a SQL table, view, or query into a pandas DataFrame.

    Parameters
    ----------
    table_name : str
        Name of SQL table or view, without the schema, or a query string
    creds : bcpandas.SqlCreds
        The credentials used in the SQL database.
    sql_type : {'table', 'view', 'query'}, default 'table'
        The type of SQL object that the parameter `table_name` is.
    schema : str, default 'dbo'
        The SQL schema of the table or view. If a query, will be ignored.
    batch_size : int, optional
        Rows will be read in batches of this size at a time. By default,
        all rows will be read at once.
    debug : bool, default False
        If True, will not delete the temporary CSV file, and will output its location.

    Returns
    -------
    `pandas.DataFrame`

    Notes
    -----
    Will actually read the SQL table/view/query twice - first using the sqlcmd utility 
    to get the names of the columns (will only read first few rows), then all the rows 
    using BCP.
    """
    # check params
    assert sql_type in SQL_TYPES

    # set up objects
    if ";" in table_name:
        raise BCPandasValueError(
            "The SQL item cannot contain the ';' character, it interferes with getting the column names"
        )

    # read top 2 rows of query to get the columns
    logger.debug("Starting to read first 2 rows using sqlcmd, to get the column names")
    _from_clause = table_name if sql_type in (TABLE, VIEW) else f"({table_name})"
    _existing_data = sqlcmd(creds=creds, command=f"SELECT TOP 2 * FROM {_from_clause} as qry")
    if _existing_data is not None:
        cols = _existing_data.columns
        logger.debug("Successfully read the column names using sqlcmd")
    else:
        raise BCPandasValueError(
            f"No data returned from the SQL item named {table_name} with type of {sql_type}"
        )

    file_path = get_temp_file()
    try:
        bcp(
            sql_item=table_name,
            direction=QUERYOUT if sql_type == QUERY else OUT,
            flat_file=file_path,
            creds=creds,
            sql_type=sql_type,
            schema=schema,
            batch_size=batch_size,
        )
        logger.debug(f"Saved dataframe to temp CSV file at {file_path}")
        return pd.read_csv(
            filepath_or_buffer=file_path,
            sep=_DELIMITER_OPTIONS[0],
            header=None,
            names=cols,
            index_col=False,
        )
    finally:
        if not debug:
            logger.debug(f"Deleting temp CSV file")
            os.remove(file_path)
        else:
            logger.debug(
                f"`read_sql` DEBUG mode, not deleting the file. CSV file is at {file_path}."
            )



# -*- coding: utf-8 -*-
"""
Created on Sat Aug  3 23:07:15 2019

@author: ydima
"""

import logging
import os
import random
import string
import subprocess
import tempfile
from io import StringIO

import pandas as pd

from .constants import (
    BCPandasException,
    BCPandasValueError,
    _DELIMITER_OPTIONS,
    DIRECTIONS,
    IN,
    NEWLINE,
    OUT,
    QUERY,
    QUERYOUT,
    SQLCHAR,
    TABLE,
    sql_collation,
)

logger = logging.getLogger(__name__)


def bcp(
    sql_item,
    direction,
    flat_file,
    creds,
    sql_type="table",
    schema="dbo",
    format_file_path=None,
    batch_size=None,
):

    combos = {TABLE: [IN, OUT], QUERY: [QUERYOUT]}
    direc = direction.lower()
    # validation
    if direc not in DIRECTIONS:
        raise BCPandasValueError(
            f"Param 'direction' must be one of {DIRECTIONS}, you passed {direc}"
        )
    if direc not in combos[sql_type]:
        raise BCPandasValueError(
            f"Wrong combo of direction and SQL object, you passed {sql_type} and {direc} ."
        )

    # auth
    if creds.with_krb_auth:
        auth = ["-T"]
    else:
        auth = ["-U", creds.username, "-P", creds.password]

    # prepare SQL item string
    if sql_type == QUERY:
        # remove newlines for queries, otherwise messes up BCP
        sql_item_string = "".join(sql_item.splitlines())
    else:
        sql_item_string = f"{schema}.{sql_item}"

    # construct BCP command
    bcp_command = [
        "bcp",
        sql_item_string,
        direc,
        flat_file,
        "-S",
        creds.server,
        "-d",
        creds.database,
    ] + auth

    if batch_size:
        bcp_command += ["-b", str(batch_size)]

    # formats
    if direc == IN:
        bcp_command += ["-f", format_file_path]
    elif direc in (OUT, QUERYOUT):
        bcp_command += [
            "-c",  # marking as character data, not Unicode (maybe make as param?)
            f"-t{_DELIMITER_OPTIONS[0]}",  # marking the delimiter as a comma (maybe also make as param?)
        ]

    # execute
    # TODO better logging and error handling the return stream
    bcp_command_log = [c if c != creds.password else "[REDACTED]" for c in bcp_command]
    logger.info(f"Executing BCP command now... \nBCP command is: {bcp_command_log}")
    result = subprocess.run(
        bcp_command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding="utf-8"
    )
    logger.debug(result.stdout)
    if result.returncode:
        logger.error(result.stdout)
        msg = parse_subprocess_error(result)
        raise BCPandasException(f"Bcp command failed. Details:\n{msg}")


def sqlcmd(creds, command):
    """
    Runs the input command against the database and returns the output if it is a table.
    
    Parameters
    ----------
    creds : bcpandas.SqlCreds
        Creds for the database
    command : str
        SQL command to be executed against the server
    
    Returns
    --------------------
    Pandas.DataFrame or None
        Returns a table if the command has an output. Returns None if the output does not return anything.
    """
    if creds.with_krb_auth:
        auth = ["-E"]
    else:
        auth = ["-U", creds.username, "-P", creds.password]
    if '"' in command:
        raise BCPandasValueError(
            'Cannot have double quotes charachter (") in the command, '
            "raises problems when combining the sqlcmd utility with Python"
        )
    command = f"set nocount on; {command} "
    sqlcmd_command = (
        ["sqlcmd", "-S", creds.server, "-d", creds.database, "-b"]
        + auth
        + ["-s,", "-W", "-Q", command]
    )

    # execute
    # TODO better logging and error handling the return stream
    sqlcmd_command_log = [c if c != creds.password else "[REDACTED]" for c in sqlcmd_command]
    logger.info(f"Executing SqlCmd command now... \nSqlCmd command is: {sqlcmd_command_log}")
    result = subprocess.run(
        sqlcmd_command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding="utf-8"
    )
    logger.debug(result.stdout)
    if result.returncode:
        logger.error(result.stdout)
        msg = parse_subprocess_error(result)
        raise BCPandasException(f"SqlCmd command failed. Details:\n{msg}")

    output = StringIO(result.stdout)
    first_line_output = output.readline().strip()
    if first_line_output == "":
        header = None
    else:
        header = "infer"
    output.seek(0)
    try:
        result = pd.read_csv(filepath_or_buffer=output, skiprows=[1], header=header)
    except pd.errors.EmptyDataError:
        result = None
    return result


def get_temp_file():
    """
    Returns full path to a temporary file without creating it.
    """
    tmp_dir = tempfile.gettempdir()
    file_path = os.path.join(
        tmp_dir, "".join(random.choices(string.ascii_letters + string.digits, k=21))
    )
    return file_path


def _escape(input_string):
    """
    Adopted from https://github.com/titan550/bcpy/blob/master/bcpy/format_file_builder.py#L25
    """
    return (
        input_string.replace('"', '\\"')
        .replace("'", "\\'")
        .replace("\r", "\\r")
        .replace("\n", "\\n")
    )


def build_format_file(df, delimiter):
    """
    Creates the non-xml SQL format file. Puts 4 spaces between each section.
    See https://docs.microsoft.com/en-us/sql/relational-databases/import-export/non-xml-format-files-sql-server
    for the specification of the file.

    # TODO add params/options to control:
    #   - the char type (not just SQLCHAR),
    #   - the ability to skip destination columns

    Parameters
    ----------
    df : pandas DataFrame
    delimiter : a valid delimiter character

    Returns
    -------
    A string containing the format file
    """
    _space = " " * 4
    format_file_str = f"9.0\n{len(df.columns)}\n"  # Version and Number of columns
    for col_num, col_name in enumerate(df.columns, start=1):
        # last col gets a newline sep
        _delim = delimiter if col_num != len(df.columns) else NEWLINE
        _line = _space.join(
            [
                str(col_num),  # Host file field order
                SQLCHAR,  # Host file data type
                str(0),  # Prefix length
                str(0),  # Host file data length
                f'"{_escape(_delim)}"',  # Terminator (see note below)
                str(col_num),  # Server column order
                col_name,  # Server column name, optional as long as not blank
                sql_collation,  # Column collation
                "\n",
            ]
        )
        format_file_str += _line
    # FYI very important to surround the Terminator with quotes, otherwise BCP fails with:
    # "Unexpected EOF encountered in BCP data-file". Hugely frustrating bug.
    return format_file_str


def _get_sql_create_statement(df, table_name, schema="dbo"):
    """
    Creates a SQL drop and re-create statement corresponding to the columns list of the object.
    
    Parameters
    -------------
    df : pandas DataFrame
    table_name : str
        name of the new table
    
    Returns
    -------------
    SQL code to create the table
    """
    sql_cols = ",".join(map(lambda x: f"[{x}] nvarchar(max)", df.columns))
    sql_command = (
        f"if object_id('[dbo].[{table_name}]', 'U') "
        f"is not null drop table [dbo].[{table_name}];"
        f"create table [dbo].[{table_name}] ({sql_cols});"
    )
    return sql_command


def parse_subprocess_error(result):
    msg = {}
    for item in ["args", "returncode", "stdout", "stderr"]:
        _i = getattr(result, item)
        msg[item] = _i.decode() if isinstance(_i, bytes) else _i
    return msg



import json
from setuptools import setup, find_packages

with open("./dist.json", "r") as file:
    config = json.load(file)

with open("./README.md", "r") as file:
    long_description = file.read()


setup(
    name=config["name"],
    version=config["version"],
    author=config["author"],
    description=config["short_description"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=f"https://github.com/{config['GH_user']}/{config['name']}",
    packages=find_packages(exclude=["tests.*", "tests"]),
    install_requires=config["dependencies"],
    python_requires=config["python_version"],
    keywords="bcp mssql pandas",
    classifiers=[
        "Topic :: Database",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Financial and Insurance Industry",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: SQL",
    ],
)
